"""Polynomials."""

from symfem.polynomials.dual import l2_dual
from symfem.polynomials.jacobi import jacobi_polynomial, monic_jacobi_polynomial
from symfem.polynomials.legendre import orthogonal_basis, orthonormal_basis
from symfem.polynomials.lobatto import lobatto_basis, lobatto_dual_basis
from symfem.polynomials.polysets import (
    Hcurl_polynomials,
    Hcurl_quolynomials,
    Hcurl_serendipity,
    Hdiv_polynomials,
    Hdiv_quolynomials,
    Hdiv_serendipity,
    polynomial_set,
    polynomial_set_1d,
    polynomial_set_vector,
    prism_polynomial_set_1d,
    prism_polynomial_set_vector,
    pyramid_polynomial_set_1d,
    pyramid_polynomial_set_vector,
    quolynomial_set_1d,
    quolynomial_set_vector,
    serendipity_indices,
    serendipity_set_1d,
    serendipity_set_vector,
)
from symfem.polynomials.degree_computation import degree
